"""
    QuaO Project backend_holder.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ...backend.backend_information import BackendInformation


class BackendDataHolder:
    def __init__(self,
                 backend_information: BackendInformation,
                 user_token: str):
        self.backend_information = backend_information
        self.user_token = user_token



"""
    QuaO Project post_processing_promise.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..callback.callback_url import CallbackUrl
from ..response.authentication import Authentication
from ..response.project_header import ProjectHeader
from ...data.promise.promise import Promise


class PostProcessingPromise(Promise):
    def __init__(self,
                 callback_url: CallbackUrl,
                 authentication: Authentication,
                 job_result,
                 project_header: ProjectHeader):
        super().__init__(callback_url, authentication)
        self.job_result = job_result
        self.project_header = project_header



"""
    QuaO Project qiskit_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import ABC

from qiskit import QiskitError

from ...model.device.device import Device
from ...config.logging_config import logger


class QiskitDevice(Device, ABC):

    def _produce_histogram_data(self, job_result) -> dict | None:
        logger.debug('[Qiskit] Produce histogram')

        try:
            histogram_data = job_result.get_counts()
        except QiskitError as qiskit_error:
            logger.debug("Can't produce histogram with error: {0}".format(str(qiskit_error)))
            histogram_data = None

        return histogram_data

    def _get_provider_job_id(self, job) -> str:
        logger.debug('[Qiskit] Get provider job id')

        return job.job_id()

    def _get_job_status(self, job) -> str:
        logger.debug('[Qiskit] Get job status')

        return job.status().name

    def _calculate_execution_time(self, job_result):
        logger.debug('[Qiskit] Calculate execution time')

        if "metadata" not in job_result:
            return None

        metadata = job_result["metadata"]

        if metadata is None or not bool(metadata) or "time_taken_execute" not in metadata:
            return None

        self.execution_time = metadata["time_taken_execute"]

        logger.debug('[Qiskit] Execution time calculation was: {0} seconds'
                     .format(self.execution_time))



"""
    QuaO Project handler.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import abstractmethod


class Handler:
    def __init__(self,
                 request_data: dict,
                 post_processing_fn):
        self.request_data = request_data
        self.post_processing_fn = post_processing_fn

    @abstractmethod
    def handle(self):
        """

        """
        pass



"""
    QuaO Project dwave_system_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from dwave.system import DWaveSampler, AutoEmbeddingComposite
from dwave.cloud import Client

from ...enum.provider_tag import ProviderTag
from ...model.provider.provider import Provider
from ...config.logging_config import *

system_devices = ["Advantage_system4.1"]
hybrid_devices = ["hybrid_binary_quadratic_model_version2"]


class DwaveSystemProvider(Provider):

    def __init__(self, api_token, endpoint):
        super().__init__(ProviderTag.D_WAVE)
        self.api_token = api_token
        self.endpoint = endpoint

    def get_backend(self, device_specification: str):
        logger.debug("[Dwave system] Get backend")

        if device_specification in system_devices:
            provider = self.collect_provider()
            return AutoEmbeddingComposite(provider)

        if device_specification in hybrid_devices:
            client = Client(endpoint=self.endpoint, token=self.api_token)
            return client.get_solver(device_specification)

        raise Exception("Unsupported Dwave device: {0}".format(device_specification))

    def collect_provider(self):
        logger.debug("[Dwave system] Connect to provider")

        return DWaveSampler(endpoint=self.endpoint, token=self.api_token)



"""
    QuaO Project base_enum.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ..enum.base_enum import BaseEnum


class ProviderTag(BaseEnum):
    QUAO_QUANTUM_SIMULATOR = 'QUAO_QUANTUM_SIMULATOR'
    IBM_QUANTUM = 'IBM_QUANTUM'
    IBM_CLOUD = 'IBM_CLOUD'
    AWS_BRAKET = 'AWS_BRAKET'
    D_WAVE = 'D_WAVE'

    @staticmethod
    def resolve(provider_tag: str):
        for element in ProviderTag:
            if element.value.__eq__(provider_tag):
                return element

        raise Exception("Provider tag is not supported!")



"""
    QuaO Project base_enum.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from enum import Enum


class BaseEnum(Enum):
    def __eq__(self, other):
        return self.name == other.name and self.value == other.value

    def __hash__(self):
        return super().__hash__()

    @staticmethod
    def resolve(type: str):
        pass



"""
    QuaO Project project_header.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""


class ProjectHeader:
    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value



"""
    QuaO Project authentication.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""


class Authentication:
    def __init__(self, user_token: str, user_identity: str):
        self.user_token = user_token
        self.user_identity = user_identity



"""
    QuaO Project json_parser_utils.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from datetime import datetime
from enum import Enum

import numpy as np


class JsonParserUtils:

    @staticmethod
    def parse(unparsed_input):
        if isinstance(unparsed_input, (complex, datetime)):
            return unparsed_input.__str__()

        if isinstance(unparsed_input, (np.int64, np.int32)):
            return int(unparsed_input)

        if isinstance(unparsed_input, (Enum,)):
            return str(unparsed_input)

        if isinstance(unparsed_input, (str, int, float, bool,)) or unparsed_input is None:
            return unparsed_input

        if isinstance(unparsed_input, (set, tuple)):
            return JsonParserUtils.parse(list(unparsed_input))

        if isinstance(unparsed_input, np.ndarray):
            return JsonParserUtils.parse(unparsed_input.tolist())

        if isinstance(unparsed_input, dict):
            data_holder = {}
            for key, value in unparsed_input.items():
                data_holder[key] = JsonParserUtils.parse(value)
            return data_holder

        if isinstance(unparsed_input, (list,)):
            data_list_holder = []
            for data in unparsed_input:
                data = JsonParserUtils.__resolve_type(data)
                data_list_holder.append(JsonParserUtils.parse(data))
            return data_list_holder

        dir_obj = dir(unparsed_input)

        if dir_obj.__contains__('to_dict'):
            return JsonParserUtils.parse(unparsed_input.to_dict())

        if dir_obj.__contains__('__dict__'):
            return JsonParserUtils.parse(unparsed_input.__dict__)

        return unparsed_input.__str__()

    @staticmethod
    def __resolve_type(element):
        if isinstance(element, (complex, datetime)):
            return element.__str__()

        if isinstance(element, (np.int64, np.int32)):
            return int(element)

        if isinstance(element, (dict, str, int, float, bool, Enum)) or element is None:
            return element

        if isinstance(element, np.ndarray):
            return JsonParserUtils.__resolve_type(element.tolist())

        if isinstance(element, (list, tuple,)):
            data_holder = []
            for data in element:
                data_holder.append(JsonParserUtils.__resolve_type(data))
            return data_holder

        return element.__dict__



"""
    QuaO Project quao_dwave_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

import time
import uuid

from ...data.device.circuit_running_option import CircuitRunningOption
from ...enum.status.job_status import JobStatus
from ...model.device.quao_device import QuaoDevice
from ...config.logging_config import logger


class QuaoDwaveDevice(QuaoDevice):
    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug('[Quao Annealing] Create job with {0} shots'.format(options.shots))

        start_time = time.time()

        job = self.device.sample(circuit)

        self.execution_time = time.time() - start_time

        return job

    def _produce_histogram_data(self, job_result) -> dict | None:
        logger.debug('[Quao Annealing] Produce histogram')

        return None

    def _get_provider_job_id(self, job) -> str:
        logger.debug('[Quao Annealing] Get provider job id')

        return str(uuid.uuid4())

    def _get_job_status(self, job) -> str:
        logger.debug('[Quao Annealing] Get job status')

        return JobStatus.DONE.value

    def _get_job_result(self, job):
        logger.debug('[Quao Annealing] Get job result')

        return job



"""
    QuaO Project circuit_holder.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""


class CircuitDataHolder:
    def __init__(self,
                 circuit,
                 export_url: str):
        self.circuit = circuit
        self.export_url = export_url



"""
    QuaO Project export_circuit.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from io import BytesIO

import requests
from braket.circuits import Circuit
from qbraid import circuit_wrapper
from qiskit import transpile

from .async_task import AsyncTask
from ..data.async_task.circuit_export.backend_holder import BackendDataHolder
from ..data.async_task.circuit_export.circuit_holder import CircuitDataHolder
from ..config.logging_config import logger
from ..enum.media_type import MediaType
from ..enum.provider_tag import ProviderTag
from ..enum.sdk import Sdk
from ..factory.provider_factory import ProviderFactory
from ..util.file_utils import FileUtils
from ..util.http_utils import HttpUtils


class CircuitExportTask(AsyncTask):
    MAX_CIRCUIT_IMAGE_SIZE = 5 * (1024 ** 2)

    def __init__(self,
                 circuit_data_holder: CircuitDataHolder,
                 backend_data_holder: BackendDataHolder):
        super().__init__()
        self.circuit_data_holder = circuit_data_holder
        self.backend_data_holder = backend_data_holder

    def do(self):
        """
          Export circuit to svg file then send to QuaO server for saving
        """
        logger.debug("[Circuit export] Start")

        circuit_export_url = self.circuit_data_holder.export_url

        if circuit_export_url is None or len(circuit_export_url) < 1:
            return

        figure_buffer = self.__convert()

        io_buffer_value, content_type = self.__determine_zip(figure_buffer=figure_buffer)

        self.__send(io_buffer_value=io_buffer_value,
                    content_type=content_type)

    def __convert(self):
        """

        @return:
        """
        logger.debug("[Circuit export] Preparing circuit figure...")
        transpiled_circuit = self.__transpile_circuit()
        circuit_figure = transpiled_circuit.draw(output='mpl', fold=-1)

        logger.debug("[Circuit export] Converting circuit figure to svg file...")
        figure_buffer = BytesIO()
        circuit_figure.savefig(figure_buffer, format='svg', bbox_inches='tight')

        return figure_buffer

    @staticmethod
    def __determine_zip(figure_buffer):
        """

        @param figure_buffer:
        @return:
        """
        buffer_value = figure_buffer.getvalue()
        content_type = MediaType.SVG_XML

        logger.debug("[Circuit export] Checking max file size")
        estimated_file_size = len(buffer_value)

        if estimated_file_size > CircuitExportTask.MAX_CIRCUIT_IMAGE_SIZE:
            logger.debug("[Circuit export] Zip file")
            zip_file_buffer = FileUtils.zip(io_buffer_value=buffer_value,
                                            file_name="circuit_image.svg")

            buffer_value = zip_file_buffer.getvalue()
            content_type = MediaType.APPLICATION_ZIP

        return buffer_value, content_type

    def __send(self, io_buffer_value, content_type: MediaType):
        """

        @param io_buffer_value:
        @param content_type:
        """
        url = self.circuit_data_holder.export_url

        logger.debug(
            "[Circuit export] Sending circuit svg image to [{0}] with POST method ...".format(
                url))

        payload = {'circuit': (
            'circuit_image.svg',
            io_buffer_value,
            content_type.value)}

        response = requests.post(
            url=url,
            headers=HttpUtils.create_bearer_header(self.backend_data_holder.user_token),
            files=payload)

        if response.ok:
            logger.debug("Sending request to QuaO backend successfully!")
        else:
            logger.debug("Sending request to QuaO backend failed with status {0}!".format(
                response.status_code))

        logger.debug("[Circuit export] Finish")

    def __transpile_circuit(self):
        """

        @return: Transpiled circuit
        """
        circuit = self.circuit_data_holder.circuit
        logger.debug("[Circuit export] Transpile circuit")

        if isinstance(circuit, Circuit):
            return circuit_wrapper(circuit).transpile(Sdk.QISKIT.value)

        provider_tag = self.backend_data_holder.backend_information.provider_tag

        if ProviderTag.AWS_BRAKET.__eq__(provider_tag):
            provider_tag = ProviderTag.QUAO_QUANTUM_SIMULATOR

        provider = ProviderFactory.create_provider(
            provider_type=provider_tag,
            sdk=Sdk.QISKIT,
            authentication=self.backend_data_holder.backend_information.authentication)

        backend = provider.get_backend(self.backend_data_holder.backend_information.device_name)

        return transpile(circuits=circuit, backend=backend)




"""
    QuaO Project quao_dwave_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from dwave.samplers import SimulatedAnnealingSampler

from ...config.logging_config import logger
from ...model.provider.quao_provider import QuaoProvider


class QuaoDwaveProvider(QuaoProvider):

    def get_backend(self, device_specification):
        logger.debug('[Quao Dwave] Get backend')

        return SimulatedAnnealingSampler()

    def collect_provider(self):
        return None



"""
    QuaO Project promise.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..callback.callback_url import CallbackUrl
from ...data.response.authentication import Authentication


class Promise:
    def __init__(self, callback_url: CallbackUrl, authentication: Authentication):
        self.callback_url = callback_url
        self.authentication = authentication



"""
    QuaO Project dwave_system_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ..device.dwave_device import DwaveDevice
from ...data.device.circuit_running_option import CircuitRunningOption
from ...config.logging_config import *


class DwaveSystemDevice(DwaveDevice):

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug("[Dwave System] Create job")

        return self.device.sample(circuit)
